
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.14.1.post1',
    'branch_nick': '(detached',
    'revision_id': '8a0d8a19502be4e8fa576196f321a6e71d0dd2dd',
    'revno': 5678
}
